//
// Separable convolution of an image. 
//
// John Robinson, 1995, 1997
//
// To convolve an image with a separable point spread function, where the
// horizontal component is given by (h1 h2 .. hm), the vertical component is
// given by (v1 v2 .. vn), and the result is to be divided by k (scaling), use
// sepconv inpic outpic h h1 h2 .. hm v v1 v2 .. vn / k 
// where the h, v and / are the literal characters 'h' 'v' and '/' used to
// delimit the horiz values, the vert values, and the divisor.
// All component values must be integers.
// Example, to convolve image with line detector:
// -0.5   1   -0.5
// -1     2   -1
// -0.5   1   -0.5
// use
// sepconv inpic outpic h -1 2 -1 v 1 2 1 / 2
//
// When doing 1D filtering, be sure to convolve with k in the unfiltered
// direction.
// The subtle part of this is the symmetric extension of the pictures at 
// the boundaries. To avoid edge effects, when convolving a picture whose
// left hand side is
// a b c d e f ...
// with a filter with an odd number of taps, the picture should be 
// 'reflected' in the boundary as:
// ... f e d c b | a b c d e f ...
// where | represents the boundary.
// Convolving with a filter with an even number of taps, requires reflecting:
// ... f e d c b a | a b c d e f ...
//
// CLIP maintains a border of 32 pels around all its images, so can reflect
// the picture in each direction up to the required amount (half the length
// of the appropriate filter.
//
#include <stdlib.h>
#include "picture.h"
static int *mask;       // Points to array of filter coeffs
static int masksize;    // Num of filter coeffs (horiz + vert)
static int vstart = 0;  // Index in filter coeffs array where vert coeffs start
static int scaling;	// Filter scaling factor

void symmetric_extension(picture_of_int& pic)
// Extends picture beyond boundaries to prevent edge effects
	{
	int hlength = vstart;
	int heven = ((hlength>>1)<<1 == hlength); // = 1 if even num of taps
	int vlength = masksize - vstart;
	int veven = ((vlength>>1)<<1 == vlength); // = 1 if even num of taps
	for (int i = 0; i < pic.nrows(); i++) {
		for (int j = -hlength/2; j < 0; j++) {
			pic[i][j] = pic[i][-j-heven];
			pic[i][pic.ncols()-1-j] = pic[i][pic.ncols()-1+j+heven];
			}
		}
	for (int j = 0; j < pic.ncols(); j++) {
		for (int i = -vlength/2; i < 0; i++) {
			pic[i][j] = pic[-i-veven][j];
			pic[pic.nrows()-1-i][j] = pic[pic.nrows()-1+i+veven][j];
			}
		}
	}	
	
int hconv(int& pel, const int **buf)
// Callback to do horizontal convolution
	{
	int *p = mask;
	const int *q = &(buf[0][-vstart/2]);
	int retval = 0;
	for (int i = 0; i < vstart; i++)
		retval += *p++ * *q++;
	pel = retval;
	return 0;
	}
int vconv(int& pel, const int **buf)
// Callback to do vertical convolution
	{
	int *p = mask+vstart;
	const int **qq = &(buf[-(masksize-vstart)/2]);
	int retval = 0;
	for (int i = vstart; i < masksize; i++)
		retval += *p++ * **qq++;
	pel = retval;
	return 0;
	}
int round_and_divide(int& pel) {
// Callback to do division by scaling factor
	if (pel < 0)
		pel -= scaling/2;
	else
		pel += scaling/2;
	pel /= scaling;
	return 1;
	}

int main(int argc, char *argv[])
	{
	if (argc <= 3)
		{
		cerr << 
"Usage: sepconv inpic outpic \'h\' h1 h2 .. hm \'v\' v1 v2 .. vn \'/\' k\n\
The \'h\' and \'v\' are char literal delimiters for the start of the \n\
horizontal and vertical components of the impulse response numerator.\n\
The \'/\' is a char literal delimiter for the impulse response demoninator.\n";
		return -1;
		}
	if (argv[3][0] != 'h')
		{
cerr << "Third argument should be char literal \'h\' to delimit start of horiz \
vals\n";
		return -1;
		}
	masksize = argc-7;
	mask = new int[masksize+1];
	// Should not use last 2 positions, but allocate enough in case user
	// does not include a 'v' and a '/'
	int i, j;
	for (i = 4, j = 0; i < argc-2; i++, j++)
		{
		if (argv[i][0] == 'v')
			{
			vstart = j--;
			continue;
			}
		mask[j] = atoi(argv[i]);
		}
	if (vstart == 0)
		{
cerr << "Missing char literal \'v\' argument to delimit start of vert vals\n";
		return -1;
		}
	if (argv[i][0] != '/') {
cerr << "Missing char literal \'/\' argument to delimit scaling factor\n";
		return -1;
		}
	scaling = atoi(argv[i+1]);
	picture_of_int in(argv[1]);
	symmetric_extension(in);
	picture_of_int mid(in,hconv);//Create mid by operating on in with hconv
	symmetric_extension(mid);
	picture_of_int out(mid,vconv);
	out.point(round_and_divide); //Scale out down
	out.write(argv[2]);
	delete [] mask;
	return 1;
	}

